@@ -15,7 +15,9 @@ use tracing::{instrument, trace, warn};
 
 use crate::core::effects::{Effects, OperationType};
 use crate::core::eventlog::{CommitActivityStatus, EventCursor, EventReplayer};
-use crate::git::{Commit, MaybeZeroOid, NonZeroOid, Repo, RepoReferencesSnapshot};
+use crate::git::{Commit, MaybeZeroOid, NonZeroOid, Repo, Time};
+
+use super::repo_ext::RepoReferencesSnapshot;
 
 impl From<NonZeroOid> for eden_dag::VertexName {
     fn from(oid: NonZeroOid) -> Self {
@@ -23,13 +25,23 @@ impl From<NonZeroOid> for eden_dag::VertexName {
     }
 }
 
-impl TryFrom<eden_dag::VertexName> for NonZeroOid {
+impl TryFrom<eden_dag::VertexName> for MaybeZeroOid {
     type Error = eyre::Error;
 
     fn try_from(value: eden_dag::VertexName) -> Result<Self, Self::Error> {
         let oid = git2::Oid::from_bytes(value.as_ref())?;
         let oid = MaybeZeroOid::from(oid);
-        NonZeroOid::try_from(oid)
+        Ok(oid)
+    }
+}
+
+impl TryFrom<eden_dag::VertexName> for NonZeroOid {
+    type Error = eyre::Error;
+
+    fn try_from(value: eden_dag::VertexName) -> Result<Self, Self::Error> {
+        let oid = MaybeZeroOid::try_from(value)?;
+        let oid = NonZeroOid::try_from(oid)?;
+        Ok(oid)
     }
 }
 
@@ -59,7 +71,7 @@ impl FromIterator<NonZeroOid> for CommitSet {
 
 /// Eagerly convert a `CommitSet` into a `Vec<NonZeroOid>` by iterating over it.
 #[instrument]
-pub fn commit_set_to_vec(commit_set: &CommitSet) -> eyre::Result<Vec<NonZeroOid>> {
+pub fn commit_set_to_vec_unsorted(commit_set: &CommitSet) -> eyre::Result<Vec<NonZeroOid>> {
     let mut result = Vec::new();
     for vertex in commit_set.iter().wrap_err("Iterating commit set")? {
         let vertex = vertex.wrap_err("Evaluating vertex")?;
@@ -70,6 +82,13 @@ pub fn commit_set_to_vec(commit_set: &CommitSet) -> eyre::Result<Vec<NonZeroOid>
     Ok(result)
 }
 
+/// Union together a list of [CommitSet]s.
+pub fn union_all(commits: &[CommitSet]) -> CommitSet {
+    commits
+        .iter()
+        .fold(CommitSet::empty(), |acc, elem| acc.union(elem))
+}
+
 /// Interface to access the directed acyclic graph (DAG) representing Git's
 /// commit graph. Based on the Eden SCM DAG.
 pub struct Dag {
@@ -137,43 +156,29 @@ impl Dag {
             branch_oid_to_names,
         } = references_snapshot;
 
-        let obsolete_commits = CommitSet::from_iter(
-            observed_commits
-                .iter()
-                .copied()
-                .filter(|commit_oid| {
-                    match event_replayer
-                        .get_cursor_commit_activity_status(event_cursor, *commit_oid)
-                    {
-                        CommitActivityStatus::Active | CommitActivityStatus::Inactive => false,
-                        CommitActivityStatus::Obsolete => true,
-                    }
-                })
-                .map(CommitVertex::from)
-                .map(Ok)
-                .collect_vec(),
-        );
+        let obsolete_commits: CommitSet = observed_commits
+            .iter()
+            .copied()
+            .filter(|commit_oid| {
+                match event_replayer.get_cursor_commit_activity_status(event_cursor, *commit_oid) {
+                    CommitActivityStatus::Active | CommitActivityStatus::Inactive => false,
+                    CommitActivityStatus::Obsolete => true,
+                }
+            })
+            .collect();
 
         let dag_dir = repo.get_dag_dir();
         std::fs::create_dir_all(&dag_dir).wrap_err("Creating .git/branchless/dag dir")?;
         let dag = eden_dag::Dag::open(&dag_dir)
             .wrap_err_with(|| format!("Opening DAG directory at: {:?}", &dag_dir))?;
 
-        let observed_commits =
-            CommitSet::from_iter(observed_commits.into_iter().map(CommitVertex::from).map(Ok));
+        let observed_commits: CommitSet = observed_commits.into_iter().collect();
         let head_commit = match head_oid {
             Some(head_oid) => CommitSet::from(*head_oid),
             None => CommitSet::empty(),
         };
         let main_branch_commit = CommitSet::from(*main_branch_oid);
-        let branch_commits = CommitSet::from_iter(
-            branch_oid_to_names
-                .keys()
-                .copied()
-                .map(CommitVertex::from)
-                .map(Ok)
-                .collect_vec(),
-        );
+        let branch_commits: CommitSet = branch_oid_to_names.keys().copied().collect();
 
         Ok(Self {
             inner: dag,
@@ -186,7 +191,8 @@ impl Dag {
     }
 
     /// This function's code adapted from `GitDag`, licensed under GPL-2.
-    fn sync(&mut self, effects: &Effects, repo: &Repo) -> eden_dag::Result<()> {
+    #[instrument]
+    fn sync(&mut self, effects: &Effects, repo: &Repo) -> eyre::Result<()> {
         let master_heads = self.main_branch_commit.clone();
         let non_master_heads = self
             .observed_commits
@@ -196,13 +202,14 @@ impl Dag {
     }
 
     /// Update the DAG with the given heads.
+    #[instrument]
     pub fn sync_from_oids(
         &mut self,
         effects: &Effects,
         repo: &Repo,
         master_heads: CommitSet,
         non_master_heads: CommitSet,
-    ) -> eden_dag::Result<()> {
+    ) -> eyre::Result<()> {
         let (effects, _progress) = effects.start_operation(OperationType::UpdateCommitGraph);
         let _effects = effects;
 
@@ -301,6 +308,18 @@ impl Dag {
         }
     }
 
+    /// Get the parent OID for the given OID. Returns an error if the given OID
+    /// does not have exactly 1 parent.
+    #[instrument]
+    pub fn get_only_parent_oid(&self, oid: NonZeroOid) -> eyre::Result<NonZeroOid> {
+        let parents: CommitSet = self.inner.parents(CommitSet::from(oid))?;
+        match commit_set_to_vec_unsorted(&parents)?[..] {
+            [oid] => Ok(oid),
+            [] => Err(eyre::eyre!("Commit {} has no parents.", oid)),
+            _ => Err(eyre::eyre!("Commit {} has more than 1 parents.", oid)),
+        }
+    }
+
     /// Get the range of OIDs from `parent_oid` to `child_oid`. Note that there
     /// may be more than one path; in that case, the OIDs are returned in a
     /// topologically-sorted order.
@@ -339,7 +358,7 @@ impl Dag {
 
     /// Conduct an arbitrary query against the DAG.
     pub fn query(&self) -> &eden_dag::Dag {
-        &*self.inner.borrow()
+        self.inner.borrow()
     }
 
     /// Return the set of commits which are public (checked into the main branch).
@@ -402,6 +421,51 @@ impl Dag {
         };
         Ok(Some(path))
     }
+
+    /// Given a CommitSet, return a list of CommitSets, each representing a
+    /// connected component of the set.
+    ///
+    /// For example, if the DAG contains commits A-B-C-D-E-F and the given
+    /// CommitSet contains `B, C, E`, this will return 2 `CommitSet`s: 1
+    /// containing `B, C` and another containing only `E`
+    #[instrument]
+    pub fn get_connected_components(&self, commit_set: &CommitSet) -> eyre::Result<Vec<CommitSet>> {
+        let mut components: Vec<CommitSet> = Vec::new();
+        let mut component = CommitSet::empty();
+        let mut commits_to_connect = commit_set.clone();
+
+        // FIXME: O(n^2) algorithm (
+        // FMI see https://github.com/arxanas/git-branchless/pull/450#issuecomment-1188391763
+        for commit in commit_set_to_vec_unsorted(commit_set)? {
+            if commits_to_connect.is_empty()? {
+                break;
+            }
+
+            if !commits_to_connect.contains(&commit.into())? {
+                continue;
+            }
+
+            let mut commits = CommitSet::from(commit);
+            while !commits.is_empty()? {
+                component = component.union(&commits);
+                commits_to_connect = commits_to_connect.difference(&commits);
+
+                let parents = self.query().parents(commits.clone())?;
+                let children = self.query().children(commits.clone())?;
+                commits = parents.union(&children).intersection(&commits_to_connect);
+            }
+
+            components.push(component);
+            component = CommitSet::empty();
+        }
+
+        let connected_commits = union_all(&components);
+        assert_eq!(commit_set.count()?, connected_commits.count()?);
+        let connected_commits = commit_set.intersection(&connected_commits);
+        assert_eq!(commit_set.count()?, connected_commits.count()?);
+
+        Ok(components)
+    }
 }
 
 impl std::fmt::Debug for Dag {
@@ -414,12 +478,21 @@ impl std::fmt::Debug for Dag {
 /// being unorderable, sort them using a deterministic tie-breaking function.
 /// Commits which have been garbage collected and are no longer available in the
 /// repository are omitted.
-pub fn sort_commit_set<'repo>(
+///
+/// FIXME: this function does not use a total ordering for the sort, which could
+/// mean that it produces incorrect results. Suppose that we have a graph with
+/// parentage relationships A < B, B < C, A < D. Since D is not directly
+/// comparable with B or C, it's possible that we calculate D < B and D > C,
+/// which violates transitivity (D < B and B < C implies that D < C).
+///
+/// We only use this function to produce deterministic output, so in practice,
+/// it doesn't seem to have a serious impact.
+pub fn sorted_commit_set<'repo>(
     repo: &'repo Repo,
     dag: &Dag,
     commit_set: &CommitSet,
 ) -> eyre::Result<Vec<Commit<'repo>>> {
-    let commit_oids = commit_set_to_vec(commit_set)?;
+    let commit_oids = commit_set_to_vec_unsorted(commit_set)?;
     let mut commits: Vec<Commit> = {
         let mut commits = Vec::new();
         for commit_oid in commit_oids {
@@ -430,7 +503,7 @@ pub fn sort_commit_set<'repo>(
         commits
     };
 
-    let commit_times: HashMap<NonZeroOid, git2::Time> = commits
+    let commit_times: HashMap<NonZeroOid, Time> = commits
         .iter()
         .map(|commit| (commit.get_oid(), commit.get_time()))
         .collect();
@@ -466,55 +539,9 @@ pub fn sort_commit_set<'repo>(
             return Ordering::Greater;
         }
 
-        (commit_times[&lhs.get_oid()], lhs.get_oid())
-            .cmp(&(commit_times[&rhs.get_oid()], rhs.get_oid()))
+        (&commit_times[&lhs.get_oid()], lhs.get_oid())
+            .cmp(&(&commit_times[&rhs.get_oid()], rhs.get_oid()))
     });
 
     Ok(commits)
 }
-
-/// The result of attempting to resolve commits.
-pub enum ResolveCommitsResult<'repo> {
-    /// All commits were successfully resolved.
-    Ok {
-        /// The commits.
-        commits: Vec<Commit<'repo>>,
-    },
-
-    /// The first commit which couldn't be resolved.
-    CommitNotFound {
-        /// The identifier of the commit, as provided by the user.
-        commit: String,
-    },
-}
-
-/// Parse strings which refer to commits, such as:
-///
-/// - Full OIDs.
-/// - Short OIDs.
-/// - Reference names.
-#[instrument]
-pub fn resolve_commits<'repo>(
-    effects: &Effects,
-    repo: &'repo Repo,
-    dag: &mut Dag,
-    hashes: Vec<String>,
-) -> eyre::Result<ResolveCommitsResult<'repo>> {
-    let mut commits = Vec::new();
-    for hash in hashes {
-        let commit = match repo.revparse_single_commit(&hash)? {
-            Some(commit) => commit,
-            None => return Ok(ResolveCommitsResult::CommitNotFound { commit: hash }),
-        };
-        commits.push(commit)
-    }
-
-    let commit_oids = commits.iter().map(|commit| commit.get_oid()).collect_vec();
-    dag.sync_from_oids(
-        effects,
-        repo,
-        CommitSet::empty(),
-        CommitSet::from_iter(commit_oids.into_iter().map(CommitVertex::from).map(Ok)),
-    )?;
-    Ok(ResolveCommitsResult::Ok { commits })
-}
