@@ -24,6 +24,7 @@ pub enum OperationType {
     CheckForCycles,
     ConstrainCommits,
     DetectDuplicateCommits,
+    EvaluateRevset(Arc<String>),
     FilterByTouchedPaths,
     FilterCommits,
     FindPathToMergeBase,
@@ -33,9 +34,13 @@ pub enum OperationType {
     InitializeRebase,
     MakeGraph,
     ProcessEvents,
+    QueryWorkingCopy,
     ReadingFromCache,
     RebaseCommits,
+    RepairBranches,
+    RepairCommits,
     RunGitCommand(Arc<String>),
+    SortCommits,
     SyncCommits,
     UpdateCommitGraph,
     WalkCommits,
@@ -50,6 +55,9 @@ impl ToString for OperationType {
             OperationType::CheckForCycles => "Checking for cycles",
             OperationType::ConstrainCommits => "Creating commit constraints",
             OperationType::DetectDuplicateCommits => "Checking for duplicate commits",
+            OperationType::EvaluateRevset(revset) => {
+                return format!("Evaluating revset: {}", revset);
+            }
             OperationType::FilterByTouchedPaths => "Filtering upstream commits by touched paths",
             OperationType::FilterCommits => "Filtering commits",
             OperationType::FindPathToMergeBase => "Finding path to merge-base",
@@ -59,11 +67,15 @@ impl ToString for OperationType {
             OperationType::InitializeRebase => "Initializing rebase",
             OperationType::MakeGraph => "Examining local history",
             OperationType::ProcessEvents => "Processing events",
+            OperationType::QueryWorkingCopy => "Querying the working copy",
             OperationType::ReadingFromCache => "Reading from cache",
             OperationType::RebaseCommits => "Rebasing commits",
+            OperationType::RepairBranches => "Checking for broken branches",
+            OperationType::RepairCommits => "Checking for broken commits",
             OperationType::RunGitCommand(command) => {
                 return format!("Running Git command: {}", &command)
             }
+            OperationType::SortCommits => "Sorting commits",
             OperationType::SyncCommits => "Syncing commit stacks",
             OperationType::UpdateCommitGraph => "Updating commit graph",
             OperationType::WalkCommits => "Walking commits",
@@ -299,18 +311,21 @@ impl OperationState {
     pub fn tick(&self, nesting_level: usize) {
         lazy_static! {
             static ref CHECKMARK: String = console::style("✓").green().to_string();
-            static ref IN_PROGRESS_SPINNER_STYLE: ProgressStyle =
-                ProgressStyle::default_spinner().template("{prefix}{spinner} {wide_msg}").unwrap();
-            static ref IN_PROGRESS_BAR_STYLE: ProgressStyle =
+            static ref IN_PROGRESS_SPINNER_STYLE: Arc<Mutex<ProgressStyle>> =
+                Arc::new(Mutex::new(ProgressStyle::default_spinner().template("{prefix}{spinner} {wide_msg}").unwrap()));
+            static ref IN_PROGRESS_BAR_STYLE: Arc<Mutex<ProgressStyle>> =
                 // indicatif places the cursor at the end of the line, which may
                 // be visible in the terminal, so we add a space at the end of
                 // the line so that the length number isn't overlapped by the
                 // cursor.
-                ProgressStyle::default_bar().template("{prefix}{spinner} {wide_msg} {bar} {pos}/{len} ").unwrap();
-            static ref FINISHED_PROGRESS_STYLE: ProgressStyle = IN_PROGRESS_SPINNER_STYLE
+                Arc::new(Mutex::new(ProgressStyle::default_bar().template("{prefix}{spinner} {wide_msg} {bar} {pos}/{len} ").unwrap()));
+            static ref FINISHED_PROGRESS_STYLE: Arc<Mutex<ProgressStyle>> = Arc::new(Mutex::new(IN_PROGRESS_SPINNER_STYLE
+                .clone()
+                .lock()
+                .unwrap()
                 .clone()
                 // Requires at least two tick values, so just pass the same one twice.
-                .tick_strings(&[&CHECKMARK, &CHECKMARK]);
+                .tick_strings(&[&CHECKMARK, &CHECKMARK])));
         }
 
         let elapsed_duration = match self.start_times.iter().min() {
@@ -323,9 +338,9 @@ impl OperationState {
 
         self.progress_bar
             .set_style(match (self.start_times.as_slice(), self.has_meter) {
-                ([], _) => FINISHED_PROGRESS_STYLE.clone(),
-                ([..], false) => IN_PROGRESS_SPINNER_STYLE.clone(),
-                ([..], true) => IN_PROGRESS_BAR_STYLE.clone(),
+                ([], _) => FINISHED_PROGRESS_STYLE.lock().unwrap().clone(),
+                ([..], false) => IN_PROGRESS_SPINNER_STYLE.lock().unwrap().clone(),
+                ([..], true) => IN_PROGRESS_BAR_STYLE.lock().unwrap().clone(),
             });
         // Both `set_message` and `set_prefix` implicitly call
         // `ProgressBar::tick` and force a redraw.
@@ -515,14 +530,25 @@ impl Effects {
     }
 
     fn on_notify_progress(&self, operation_key: &OperationKey, current: usize, total: usize) {
+        match self.dest {
+            OutputDest::Stdout => {}
+            OutputDest::Suppress | OutputDest::BufferForTest { .. } => return,
+        }
+
         let mut root_operation = self.root_operation.lock().unwrap();
         let operation_state = root_operation.get_or_create_child(operation_key);
         operation_state.set_progress(current, total);
     }
 
     fn on_notify_progress_inc(&self, operation_key: &OperationKey, increment: usize) {
+        match self.dest {
+            OutputDest::Stdout => {}
+            OutputDest::Suppress | OutputDest::BufferForTest { .. } => return,
+        }
+
         let mut root_operation = self.root_operation.lock().unwrap();
-        let operation_state = match root_operation.get_child(operation_key) {
+        let operation = root_operation.get_child(operation_key);
+        let operation_state = match operation {
             Some(operation_state) => operation_state,
             None => return,
         };
@@ -530,8 +556,14 @@ impl Effects {
     }
 
     fn on_set_message(&self, operation_key: &OperationKey, message: String) {
+        match self.dest {
+            OutputDest::Stdout => {}
+            OutputDest::Suppress | OutputDest::BufferForTest { .. } => return,
+        }
+
         let mut root_operation = self.root_operation.lock().unwrap();
-        let operation_state = match root_operation.get_child(operation_key) {
+        let operation = root_operation.get_child(operation_key);
+        let operation_state = match operation {
             Some(operation_state) => operation_state,
             None => return,
         };
@@ -547,7 +579,8 @@ impl Effects {
         let now = Instant::now();
         let mut root_operation = self.root_operation.lock().unwrap();
 
-        let operation_state = match root_operation.get_child(operation_key) {
+        let operation = root_operation.get_child(operation_key);
+        let operation_state = match operation {
             Some(operation_state) => operation_state,
             None => {
                 warn!("Progress operation not started");
@@ -625,7 +658,8 @@ trait WriteProgress {
         // `ProgressBar` is included in a `MultiProgress`, it doesn't matter
         // which of them we call `println` on. The output will be printed above
         // the `MultiProgress` regardless.
-        match root_operation.children.get(0) {
+        let operation = root_operation.children.get(0);
+        match operation {
             None => {
                 // There's no progress meters, so we can write directly to
                 // stdout. Note that we don't style output here; instead, we
