@@ -17,9 +17,10 @@ use eyre::Context;
 use tracing::{error, instrument};
 
 use crate::core::effects::{Effects, OperationType};
-use crate::git::{
-    CategorizedReferenceName, MaybeZeroOid, NonZeroOid, Repo, RepoReferencesSnapshot,
-};
+use crate::core::repo_ext::RepoExt;
+use crate::git::{CategorizedReferenceName, MaybeZeroOid, NonZeroOid, Repo};
+
+use super::repo_ext::RepoReferencesSnapshot;
 
 /// When this environment variable is set, we reuse the ID for the transaction
 /// which the caller has already started.
@@ -53,7 +54,7 @@ struct Row {
 ///
 /// Unlike in a database, there is no specific guarantee that an event
 /// transaction is an atomic unit of work.
-#[derive(Clone, Copy, Debug, PartialEq)]
+#[derive(Clone, Copy, Debug, PartialEq, Eq)]
 pub struct EventTransactionId(isize);
 
 impl ToString for EventTransactionId {
@@ -165,6 +166,27 @@ pub enum Event {
         /// The OID of the commit that was unobsoleted.
         commit_oid: NonZeroOid,
     },
+
+    /// Represents a snapshot of the working copy made at a certain time,
+    /// typically before a potentially-destructive operation.
+    WorkingCopySnapshot {
+        /// The timestamp of the event.
+        timestamp: f64,
+
+        /// The transaction ID of the event.
+        event_tx_id: EventTransactionId,
+
+        /// The OID of the current HEAD commit.
+        head_oid: MaybeZeroOid,
+
+        /// The OID of the commit containing metadata about the working copy
+        /// snapshot.
+        commit_oid: NonZeroOid,
+
+        /// The name of the checked-out branch, if any. This should be a full
+        /// reference name like `refs/heads/foo`.
+        ref_name: Option<OsString>,
+    },
 }
 
 impl Event {
@@ -176,6 +198,7 @@ impl Event {
             Event::CommitEvent { timestamp, .. } => timestamp,
             Event::ObsoleteEvent { timestamp, .. } => timestamp,
             Event::UnobsoleteEvent { timestamp, .. } => timestamp,
+            Event::WorkingCopySnapshot { timestamp, .. } => timestamp,
         };
         SystemTime::UNIX_EPOCH + Duration::from_secs_f64(*timestamp)
     }
@@ -188,6 +211,7 @@ impl Event {
             Event::CommitEvent { event_tx_id, .. } => *event_tx_id,
             Event::ObsoleteEvent { event_tx_id, .. } => *event_tx_id,
             Event::UnobsoleteEvent { event_tx_id, .. } => *event_tx_id,
+            Event::WorkingCopySnapshot { event_tx_id, .. } => *event_tx_id,
         }
     }
 }
@@ -268,6 +292,22 @@ impl From<Event> for Row {
                 ref_name: None,
                 message: None,
             },
+
+            Event::WorkingCopySnapshot {
+                timestamp,
+                event_tx_id: EventTransactionId(event_tx_id),
+                head_oid,
+                commit_oid,
+                ref_name,
+            } => Row {
+                timestamp,
+                event_tx_id,
+                type_: String::from("snapshot"),
+                ref1: Some(head_oid.to_string().into()),
+                ref2: Some(commit_oid.to_string().into()),
+                ref_name,
+                message: None,
+            },
         }
     }
 }
@@ -356,6 +396,18 @@ fn try_from_row_helper(row: &Row) -> Result<Event, eyre::Error> {
             }
         }
 
+        "snapshot" => {
+            let head_oid: MaybeZeroOid = get_oid(&ref1, "head OID")?;
+            let commit_oid: NonZeroOid = get_oid(&ref2, "commit OID")?.try_into()?;
+            Event::WorkingCopySnapshot {
+                timestamp,
+                event_tx_id,
+                head_oid,
+                commit_oid,
+                ref_name,
+            }
+        }
+
         other => eyre::bail!("Unknown event type {}", other),
     };
     Ok(event)
@@ -438,7 +490,7 @@ impl<'conn> EventLogDb<'conn> {
     /// Args:
     /// * events: The events to add.
     #[instrument]
-    pub fn add_events(&mut self, events: Vec<Event>) -> eyre::Result<()> {
+    pub fn add_events(&self, events: Vec<Event>) -> eyre::Result<()> {
         let tx = self.conn.unchecked_transaction()?;
         for event in events {
             let Row {
@@ -574,6 +626,26 @@ ORDER BY rowid ASC
     ) -> eyre::Result<EventTransactionId> {
         self.make_transaction_id_inner(now, message.as_ref())
     }
+
+    /// Get the message associated with the given transaction.
+    pub fn get_transaction_message(&self, event_tx_id: EventTransactionId) -> eyre::Result<String> {
+        let EventTransactionId(event_tx_id) = event_tx_id;
+        let mut stmt = self.conn.prepare(
+            "
+SELECT message
+FROM event_transactions
+WHERE event_tx_id = :event_tx_id
+",
+        )?;
+        let result: String = stmt.query_row(
+            rusqlite::named_params![":event_tx_id": event_tx_id,],
+            |row| {
+                let message: String = row.get("message")?;
+                Ok(message)
+            },
+        )?;
+        Ok(result)
+    }
 }
 
 /// Determine whether a given reference is used to keep a commit alive.
@@ -651,7 +723,7 @@ struct EventInfo {
 /// Thus, all events before to the cursor are considered to be in effect,
 /// and all events after the cursor are considered to not have happened
 /// yet.
-#[derive(Clone, Copy, Debug, PartialEq)]
+#[derive(Clone, Copy, Debug, PartialEq, Eq)]
 pub struct EventCursor {
     event_id: isize,
 }
@@ -838,6 +910,11 @@ impl EventReplayer {
                     event: event.clone(),
                     event_classification: EventClassification::Show,
                 }),
+
+            Event::WorkingCopySnapshot { .. } => {
+                // Do nothing. A working copy snapshot doesn't imply that the
+                // commit has become active or inactive.
+            }
         };
     }
 
@@ -1098,6 +1175,15 @@ impl EventReplayer {
                     // doesn't have the corresponding `RefUpdateEvent`.
                     Event::CommitEvent { commit_oid, .. } => Some(*commit_oid),
 
+                    Event::WorkingCopySnapshot {
+                        head_oid: MaybeZeroOid::NonZero(head_oid),
+                        ..
+                    } => Some(*head_oid),
+                    Event::WorkingCopySnapshot {
+                        head_oid: MaybeZeroOid::Zero,
+                        ..
+                    } => None,
+
                     Event::RewriteEvent { .. }
                     | Event::ObsoleteEvent { .. }
                     | Event::UnobsoleteEvent { .. } => None,
@@ -1286,17 +1372,20 @@ pub mod testing {
         match event {
             Event::RewriteEvent {
                 ref mut timestamp, ..
-            } => *timestamp = 0.0,
-            Event::RefUpdateEvent {
+            }
+            | Event::RefUpdateEvent {
                 ref mut timestamp, ..
-            } => *timestamp = 0.0,
-            Event::CommitEvent {
+            }
+            | Event::CommitEvent {
                 ref mut timestamp, ..
-            } => *timestamp = 0.0,
-            Event::ObsoleteEvent {
+            }
+            | Event::ObsoleteEvent {
                 ref mut timestamp, ..
-            } => *timestamp = 0.0,
-            Event::UnobsoleteEvent {
+            }
+            | Event::UnobsoleteEvent {
+                ref mut timestamp, ..
+            }
+            | Event::WorkingCopySnapshot {
                 ref mut timestamp, ..
             } => *timestamp = 0.0,
         }
