@@ -7,9 +7,8 @@ use std::borrow::Cow;
 use std::collections::{HashMap, HashSet};
 use std::convert::TryInto;
 use std::ffi::OsString;
-use std::ops::Add;
 use std::sync::{Arc, Mutex};
-use std::time::{Duration, SystemTime};
+use std::time::SystemTime;
 
 use cursive::theme::BaseColor;
 use cursive::utils::markup::StyledString;
@@ -22,13 +21,11 @@ use crate::core::config::{
     get_commit_descriptors_branches, get_commit_descriptors_differential_revision,
     get_commit_descriptors_relative_time,
 };
-use crate::git::{
-    CategorizedReferenceName, Commit, NonZeroOid, Repo, RepoReferencesSnapshot,
-    ResolvedReferenceInfo,
-};
+use crate::git::{CategorizedReferenceName, Commit, NonZeroOid, Repo, ResolvedReferenceInfo};
 
 use super::eventlog::{Event, EventCursor, EventReplayer};
 use super::formatting::{Glyphs, StyledStringBuilder};
+use super::repo_ext::RepoReferencesSnapshot;
 use super::rewrite::find_rewrite_target;
 
 /// An object which can be rendered in the smartlog.
@@ -286,7 +283,8 @@ impl<'a> NodeDescriptor for ObsolescenceExplanationDescriptor<'a> {
 
             Event::RefUpdateEvent { .. }
             | Event::CommitEvent { .. }
-            | Event::UnobsoleteEvent { .. } => None,
+            | Event::UnobsoleteEvent { .. }
+            | Event::WorkingCopySnapshot { .. } => None,
         };
         Ok(result)
     }
@@ -512,9 +510,7 @@ impl NodeDescriptor for RelativeTimeDescriptor {
             NodeObject::GarbageCollected { oid: _ } => return Ok(None),
         };
 
-        let previous_time = SystemTime::UNIX_EPOCH
-            .add(Duration::from_secs(commit.get_time().seconds().try_into()?));
-        let description = Self::describe_time_delta(self.now, previous_time)?;
+        let description = Self::describe_time_delta(self.now, commit.get_time().to_system_time()?)?;
         let result = StyledString::styled(description, BaseColor::Green.dark());
         Ok(Some(result))
     }
@@ -522,7 +518,10 @@ impl NodeDescriptor for RelativeTimeDescriptor {
 
 #[cfg(test)]
 mod tests {
-    use std::ops::Sub;
+    use std::{
+        ops::{Add, Sub},
+        time::Duration,
+    };
 
     use super::*;
 
