@@ -10,13 +10,13 @@ use eyre::Context;
 use os_str_bytes::OsStrBytes;
 use tracing::warn;
 
+use crate::core::check_out::{check_out_commit, CheckOutCommitOptions};
 use crate::core::effects::Effects;
-use crate::core::eventlog::EventTransactionId;
+use crate::core::eventlog::{EventLogDb, EventTransactionId};
 use crate::core::formatting::{printable_styled_string, Pluralize};
-use crate::git::{
-    check_out_commit, CheckOutCommitOptions, GitRunInfo, MaybeZeroOid, NonZeroOid, Repo,
-    ResolvedReferenceInfo,
-};
+use crate::core::repo_ext::RepoExt;
+use crate::git::{GitRunInfo, MaybeZeroOid, NonZeroOid, Repo, ResolvedReferenceInfo};
+use crate::util::ExitCode;
 
 use super::plan::RebasePlan;
 
@@ -138,12 +138,13 @@ pub fn check_out_updated_head(
     effects: &Effects,
     git_run_info: &GitRunInfo,
     repo: &Repo,
+    event_log_db: &EventLogDb,
     event_tx_id: EventTransactionId,
     rewritten_oids: &HashMap<NonZeroOid, MaybeZeroOid>,
     previous_head_info: &ResolvedReferenceInfo,
     skipped_head_updated_oid: Option<NonZeroOid>,
     check_out_commit_options: &CheckOutCommitOptions,
-) -> eyre::Result<isize> {
+) -> eyre::Result<ExitCode> {
     let checkout_target: ResolvedReferenceInfo = match previous_head_info {
         ResolvedReferenceInfo {
             oid: None,
@@ -235,14 +236,14 @@ pub fn check_out_updated_head(
 
     let head_info = repo.get_head_info()?;
     if head_info == checkout_target {
-        return Ok(0);
+        return Ok(ExitCode(0));
     }
 
     let checkout_target: Cow<OsStr> = match &checkout_target {
         ResolvedReferenceInfo {
             oid: None,
             reference_name: None,
-        } => return Ok(0),
+        } => return Ok(ExitCode(0)),
 
         ResolvedReferenceInfo {
             oid: Some(oid),
@@ -265,13 +266,26 @@ pub fn check_out_updated_head(
     let result = check_out_commit(
         effects,
         git_run_info,
-        Some(event_tx_id),
+        repo,
+        event_log_db,
+        event_tx_id,
         Some(&checkout_target),
         check_out_commit_options,
     )?;
     Ok(result)
 }
 
+/// What to suggest that the user do in order to resolve a merge conflict.
+#[derive(Copy, Clone, Debug)]
+pub enum MergeConflictRemediation {
+    /// Indicate that the user should retry the merge operation (but with
+    /// `--merge`).
+    Retry,
+
+    /// Indicate that the user should run `git restack --merge`.
+    Restack,
+}
+
 /// Information about a merge conflict that occurred while moving commits.
 #[derive(Debug)]
 pub struct MergeConflictInfo {
@@ -285,7 +299,12 @@ pub struct MergeConflictInfo {
 impl MergeConflictInfo {
     /// Describe the merge conflict in a user-friendly way and advise to rerun
     /// with `--merge`.
-    pub fn describe(&self, effects: &Effects, repo: &Repo) -> eyre::Result<()> {
+    pub fn describe(
+        &self,
+        effects: &Effects,
+        repo: &Repo,
+        remediation: MergeConflictRemediation,
+    ) -> eyre::Result<()> {
         writeln!(
             effects.get_output_stream(),
             "This operation would cause a merge conflict:"
@@ -304,10 +323,22 @@ impl MergeConflictInfo {
                 repo.friendly_describe_commit_from_oid(effects.get_glyphs(), self.commit_oid)?
             )?
         )?;
-        writeln!(
-            effects.get_output_stream(),
-            "To resolve merge conflicts, retry this operation with the --merge option."
-        )?;
+
+        match remediation {
+            MergeConflictRemediation::Retry => {
+                writeln!(
+                    effects.get_output_stream(),
+                    "To resolve merge conflicts, retry this operation with the --merge option."
+                )?;
+            }
+            MergeConflictRemediation::Restack => {
+                writeln!(
+                    effects.get_output_stream(),
+                    "To resolve merge conflicts, run: git restack --merge"
+                )?;
+            }
+        }
+
         Ok(())
     }
 }
@@ -321,6 +352,7 @@ mod in_memory {
     use tracing::{instrument, warn};
 
     use crate::core::effects::{Effects, OperationType};
+    use crate::core::eventlog::EventLogDb;
     use crate::core::formatting::printable_styled_string;
     use crate::core::gc::mark_commit_reachable;
     use crate::core::rewrite::execute::check_out_updated_head;
@@ -329,6 +361,7 @@ mod in_memory {
     use crate::git::{
         CherryPickFastError, CherryPickFastOptions, GitRunInfo, MaybeZeroOid, NonZeroOid, Repo,
     };
+    use crate::util::ExitCode;
 
     use super::{ExecuteRebasePlanOptions, MergeConflictInfo};
 
@@ -646,10 +679,11 @@ mod in_memory {
         effects: &Effects,
         git_run_info: &GitRunInfo,
         repo: &Repo,
+        event_log_db: &EventLogDb,
         rewritten_oids: &[(NonZeroOid, MaybeZeroOid)],
         skipped_head_updated_oid: Option<NonZeroOid>,
         options: &ExecuteRebasePlanOptions,
-    ) -> eyre::Result<isize> {
+    ) -> eyre::Result<ExitCode> {
         let ExecuteRebasePlanOptions {
             now: _,
             event_tx_id,
@@ -706,6 +740,7 @@ mod in_memory {
             effects,
             git_run_info,
             repo,
+            event_log_db,
             *event_tx_id,
             &rewritten_oids_map,
             &head_info,
@@ -729,6 +764,7 @@ mod on_disk {
     use crate::core::rewrite::plan::RebasePlan;
     use crate::core::rewrite::rewrite_hooks::save_original_head_info;
     use crate::git::{GitRunInfo, Repo};
+    use crate::util::ExitCode;
 
     use super::ExecuteRebasePlanOptions;
 
@@ -915,7 +951,7 @@ mod on_disk {
         repo: &Repo,
         rebase_plan: &RebasePlan,
         options: &ExecuteRebasePlanOptions,
-    ) -> eyre::Result<Result<isize, Error>> {
+    ) -> eyre::Result<Result<ExitCode, Error>> {
         let ExecuteRebasePlanOptions {
             // `git rebase` will make its own timestamp.
             now: _,
@@ -943,7 +979,7 @@ mod on_disk {
 
 /// Options to use when executing a `RebasePlan`.
 #[derive(Clone, Debug)]
-pub struct ExecuteRebasePlanOptions<'a> {
+pub struct ExecuteRebasePlanOptions {
     /// The time which should be recorded for this event.
     pub now: SystemTime,
 
@@ -966,7 +1002,7 @@ pub struct ExecuteRebasePlanOptions<'a> {
     pub resolve_merge_conflicts: bool,
 
     /// If `HEAD` was moved, the options for checking out the new `HEAD` commit.
-    pub check_out_commit_options: CheckOutCommitOptions<'a>,
+    pub check_out_commit_options: CheckOutCommitOptions,
 }
 
 /// The result of executing a rebase plan.
@@ -990,7 +1026,7 @@ pub enum ExecuteRebasePlanResult {
     Failed {
         /// The exit code to exit with. (This value may have been obtained from
         /// a subcommand invocation.)
-        exit_code: isize,
+        exit_code: ExitCode,
     },
 }
 
@@ -1000,6 +1036,7 @@ pub fn execute_rebase_plan(
     effects: &Effects,
     git_run_info: &GitRunInfo,
     repo: &Repo,
+    event_log_db: &EventLogDb,
     rebase_plan: &RebasePlan,
     options: &ExecuteRebasePlanOptions,
 ) -> eyre::Result<ExecuteRebasePlanResult> {
@@ -1025,14 +1062,23 @@ pub fn execute_rebase_plan(
                 rewritten_oids,
                 new_head_oid,
             } => {
-                post_rebase_in_memory(
+                // Ignore the return code, as it probably indicates that the
+                // checkout failed (which might happen if the user has changes
+                // which don't merge cleanly). The user can resolve that
+                // themselves.
+                //
+                // FIXME: we may still want to propagate the exit code to the
+                // caller.
+                let ExitCode(_exit_code) = post_rebase_in_memory(
                     effects,
                     git_run_info,
                     repo,
+                    event_log_db,
                     &rewritten_oids,
                     new_head_oid,
                     options,
                 )?;
+
                 let rewritten_oids: HashMap<NonZeroOid, MaybeZeroOid> =
                     rewritten_oids.into_iter().collect();
                 writeln!(effects.get_output_stream(), "In-memory rebase succeeded.")?;
@@ -1092,7 +1138,9 @@ pub fn execute_rebase_plan(
                 effects.get_output_stream(),
                 "Aborting since an in-memory rebase was requested."
             )?;
-            return Ok(ExecuteRebasePlanResult::Failed { exit_code: 1 });
+            return Ok(ExecuteRebasePlanResult::Failed {
+                exit_code: ExitCode(1),
+            });
         } else {
             writeln!(effects.get_output_stream(), "Trying again on-disk...")?;
         }
@@ -1101,7 +1149,7 @@ pub fn execute_rebase_plan(
     if !force_in_memory {
         use on_disk::*;
         match rebase_on_disk(effects, git_run_info, repo, rebase_plan, options)? {
-            Ok(0) => {
+            Ok(ExitCode(0)) => {
                 return Ok(ExecuteRebasePlanResult::Succeeded {
                     rewritten_oids: None,
                 });
@@ -1116,7 +1164,9 @@ in your working copy which might be overwritten as a result.
 Commit your changes and then try again.
 "
                 )?;
-                return Ok(ExecuteRebasePlanResult::Failed { exit_code: 1 });
+                return Ok(ExecuteRebasePlanResult::Failed {
+                    exit_code: ExitCode(1),
+                });
             }
             Err(Error::OperationAlreadyInProgress { operation_type }) => {
                 writeln!(
@@ -1129,7 +1179,9 @@ Commit your changes and then try again.
                     "Run git {0} --continue or git {0} --abort to resolve it and proceed.",
                     operation_type
                 )?;
-                return Ok(ExecuteRebasePlanResult::Failed { exit_code: 1 });
+                return Ok(ExecuteRebasePlanResult::Failed {
+                    exit_code: ExitCode(1),
+                });
             }
         }
     }
