@@ -8,8 +8,6 @@ use std::process::{Command, ExitStatus, Stdio};
 use std::sync::Arc;
 use std::thread::{self, JoinHandle};
 
-use cursive::theme::BaseColor;
-use cursive::utils::markup::StyledString;
 use eyre::{eyre, Context};
 use itertools::Itertools;
 use os_str_bytes::OsStrBytes;
@@ -18,11 +16,8 @@ use tracing::instrument;
 use crate::core::config::get_core_hooks_path;
 use crate::core::effects::{Effects, OperationType};
 use crate::core::eventlog::{EventTransactionId, BRANCHLESS_TRANSACTION_ID_ENV_VAR};
-use crate::core::formatting::printable_styled_string;
 use crate::git::repo::Repo;
-use crate::util::get_sh;
-
-use super::CategorizedReferenceName;
+use crate::util::{get_sh, ExitCode};
 
 /// Path to the `git` executable on disk to be executed.
 #[derive(Clone)]
@@ -47,22 +42,35 @@ impl std::fmt::Debug for GitRunInfo {
     }
 }
 
+/// Options for invoking Git.
 pub struct GitRunOpts {
     /// If set, a non-zero exit code will be treated as an error.
-    treat_git_failure_as_error: bool,
+    pub treat_git_failure_as_error: bool,
+
+    /// A vector of bytes to write to the Git process's stdin. If `None`,
+    /// nothing is written to stdin.
+    pub stdin: Option<Vec<u8>>,
 }
 
 impl Default for GitRunOpts {
     fn default() -> Self {
         Self {
             treat_git_failure_as_error: true,
+            stdin: None,
         }
     }
 }
 
+/// The result of invoking Git.
+#[must_use]
 pub struct GitRunResult {
+    /// The exit code of the process.
     pub exit_code: i32,
+
+    /// The stdout contents written by the invocation.
     pub stdout: Vec<u8>,
+
+    /// The stderr contents written by the invocation.
     pub stderr: Vec<u8>,
 }
 
@@ -105,7 +113,7 @@ impl GitRunInfo {
         effects: &Effects,
         event_tx_id: Option<EventTransactionId>,
         args: &[&OsStr],
-    ) -> eyre::Result<isize> {
+    ) -> eyre::Result<ExitCode> {
         let GitRunInfo {
             path_to_git,
             working_directory,
@@ -154,10 +162,11 @@ impl GitRunInfo {
         // On Unix, if the child process was terminated by a signal, we need to call
         // some Unix-specific functions to access the signal that terminated it. For
         // simplicity, just return `1` in those cases.
-        let exit_code = exit_status.code().unwrap_or(1);
-        let exit_code = exit_code
+        let exit_code: i32 = exit_status.code().unwrap_or(1);
+        let exit_code: isize = exit_code
             .try_into()
             .wrap_err("Converting exit code from i32 to isize")?;
+        let exit_code = ExitCode(exit_code);
         Ok(exit_code)
     }
 
@@ -171,13 +180,13 @@ impl GitRunInfo {
     ///
     /// Returns the exit code of Git (non-zero signifies error).
     #[instrument]
-    #[must_use = "The return code for `run_git` must be checked"]
+    #[must_use = "The return code for `GitRunInfo::run` must be checked"]
     pub fn run<S: AsRef<OsStr> + std::fmt::Debug>(
         &self,
         effects: &Effects,
         event_tx_id: Option<EventTransactionId>,
         args: &[S],
-    ) -> eyre::Result<isize> {
+    ) -> eyre::Result<ExitCode> {
         self.run_inner(
             effects,
             event_tx_id,
@@ -185,6 +194,47 @@ impl GitRunInfo {
         )
     }
 
+    /// Run the provided command without wrapping it in an `Effects` operation.
+    /// This may clobber progress reporting, and is usually not what you want;
+    /// see [`GitRunInfo::run`] instead.
+    #[instrument]
+    #[must_use = "The return code for `GitRunInfo::run_direct_no_wrapping` must be checked"]
+    pub fn run_direct_no_wrapping(
+        &self,
+        event_tx_id: Option<EventTransactionId>,
+        args: &[impl AsRef<OsStr> + std::fmt::Debug],
+    ) -> eyre::Result<ExitCode> {
+        let GitRunInfo {
+            path_to_git,
+            working_directory,
+            env,
+        } = self;
+
+        let mut command = Command::new(path_to_git);
+        command.current_dir(working_directory);
+        command.args(args);
+        command.env_clear();
+        command.envs(env.iter());
+        if let Some(event_tx_id) = event_tx_id {
+            command.env(BRANCHLESS_TRANSACTION_ID_ENV_VAR, event_tx_id.to_string());
+        }
+
+        let mut child = command.spawn().wrap_err("Spawning Git subprocess")?;
+        let exit_status = child
+            .wait()
+            .wrap_err("Waiting for Git subprocess to complete")?;
+
+        // On Unix, if the child process was terminated by a signal, we need to call
+        // some Unix-specific functions to access the signal that terminated it. For
+        // simplicity, just return `1` in those cases.
+        let exit_code: i32 = exit_status.code().unwrap_or(1);
+        let exit_code: isize = exit_code
+            .try_into()
+            .wrap_err("Converting exit code from i32 to isize")?;
+        let exit_code = ExitCode(exit_code);
+        Ok(exit_code)
+    }
+
     fn run_silent_inner(
         &self,
         repo: &Repo,
@@ -199,6 +249,7 @@ impl GitRunInfo {
         } = self;
         let GitRunOpts {
             treat_git_failure_as_error,
+            stdin,
         } = opts;
 
         // Prefer running in the working copy path to the repo path, because
@@ -229,7 +280,28 @@ impl GitRunInfo {
         if let Some(event_tx_id) = event_tx_id {
             command.env(BRANCHLESS_TRANSACTION_ID_ENV_VAR, event_tx_id.to_string());
         }
-        let output = command.output().wrap_err("Spawning Git subprocess")?;
+
+        command.stdin(match stdin {
+            Some(_) => Stdio::piped(),
+            None => Stdio::null(),
+        });
+        command.stdout(Stdio::piped());
+        command.stderr(Stdio::piped());
+
+        let mut child = command.spawn().wrap_err("Spawning Git subprocess")?;
+
+        if let Some(stdin) = stdin {
+            child
+                .stdin
+                .as_mut()
+                .unwrap()
+                .write_all(&stdin)
+                .wrap_err("Writing process stdin")?;
+        }
+
+        let output = child
+            .wait_with_output()
+            .wrap_err("Spawning Git subprocess")?;
         let result = GitRunResult {
             // On Unix, if the child process was terminated by a signal, we need to call
             // some Unix-specific functions to access the signal that terminated it. For
@@ -239,7 +311,11 @@ impl GitRunInfo {
             stderr: output.stderr,
         };
         if treat_git_failure_as_error && !output.status.success() {
-            eyre::bail!("Git subprocess failed: {:?}", result);
+            eyre::bail!(
+                "Git subprocess failed:\nArgs: {:?}\nResult: {:?}",
+                &args,
+                result
+            );
         }
         Ok(result)
     }
@@ -365,89 +441,8 @@ impl GitRunInfo {
     }
 }
 
-/// Options for checking out a commit.
-#[derive(Clone, Debug)]
-pub struct CheckOutCommitOptions<'a> {
-    /// Additional arguments to pass to `git checkout`.
-    pub additional_args: &'a [&'a str],
-
-    /// Whether or not to render the smartlog after the checkout has completed.
-    pub render_smartlog: bool,
-}
-
-impl Default for CheckOutCommitOptions<'_> {
-    fn default() -> Self {
-        Self {
-            additional_args: Default::default(),
-            render_smartlog: true,
-        }
-    }
-}
-
-/// Checks out the requested commit. If the operation succeeds, then displays
-/// the new smartlog. Otherwise displays a warning message.
-#[instrument]
-pub fn check_out_commit(
-    effects: &Effects,
-    git_run_info: &GitRunInfo,
-    event_tx_id: Option<EventTransactionId>,
-    target: Option<impl AsRef<OsStr> + std::fmt::Debug>,
-    options: &CheckOutCommitOptions,
-) -> eyre::Result<isize> {
-    let CheckOutCommitOptions {
-        additional_args,
-        render_smartlog,
-    } = options;
-
-    let target = match target {
-        None => None,
-        Some(target) => {
-            let categorized_target = CategorizedReferenceName::new(target.as_ref());
-            Some(categorized_target.remove_prefix()?)
-        }
-    };
-
-    let args = {
-        let mut args = vec![OsStr::new("checkout")];
-        if let Some(target) = &target {
-            args.push(target);
-        }
-        args.extend(additional_args.iter().map(OsStr::new));
-        args
-    };
-    let result = git_run_info.run(effects, event_tx_id, args.as_slice())?;
-
-    if result == 0 {
-        if *render_smartlog {
-            let result = git_run_info.run(effects, event_tx_id, &["branchless", "smartlog"])?;
-            Ok(result)
-        } else {
-            Ok(result)
-        }
-    } else {
-        writeln!(
-            effects.get_output_stream(),
-            "{}",
-            printable_styled_string(
-                effects.get_glyphs(),
-                StyledString::styled(
-                    match target {
-                        Some(target) =>
-                            format!("Failed to check out commit: {}", target.to_string_lossy()),
-                        None => "Failed to check out commit".to_string(),
-                    },
-                    BaseColor::Red.light()
-                )
-            )?
-        )?;
-        Ok(result)
-    }
-}
-
 #[cfg(test)]
 mod tests {
-    use insta::assert_debug_snapshot;
-
     use super::{GitRunInfo, GitRunOpts};
     use crate::testing::make_git;
 
@@ -511,13 +506,10 @@ mod tests {
             &["some-nonexistent-command"],
             GitRunOpts {
                 treat_git_failure_as_error: true,
+                stdin: None,
             },
         );
-        assert_debug_snapshot!(result, @r###"
-        Err(
-            "Git subprocess failed: <GitRunResult exit_code=1 stdout=\"\" stderr=\"git: 'some-nonexistent-command' is not a git command. See 'git --help'.\\n\">",
-        )
-        "###);
+        assert!(result.is_err());
 
         let result = git_run_info.run_silent(
             &git.get_repo()?,
@@ -525,13 +517,10 @@ mod tests {
             &["some-nonexistent-command"],
             GitRunOpts {
                 treat_git_failure_as_error: false,
+                stdin: None,
             },
         );
-        assert_debug_snapshot!(result, @r###"
-        Ok(
-            <GitRunResult exit_code=1 stdout="" stderr="git: 'some-nonexistent-command' is not a git command. See 'git --help'.\n">,
-        )
-        "###);
+        assert!(result.is_ok());
 
         Ok(())
     }
