@@ -9,16 +9,18 @@ use std::time::SystemTime;
 
 use eyre::Context;
 use itertools::Itertools;
+use lib::core::rewrite::MergeConflictRemediation;
+use lib::util::ExitCode;
 use tracing::instrument;
 
 use crate::commands::restack;
-use crate::opts::MoveOptions;
+use crate::opts::{MoveOptions, Revset};
 use lib::core::config::get_restack_preserve_timestamps;
 use lib::core::effects::Effects;
 use lib::core::eventlog::{Event, EventLogDb};
 use lib::core::formatting::Pluralize;
 use lib::core::gc::mark_commit_reachable;
-use lib::git::{AmendFastOptions, FileStatus, GitRunInfo, Repo};
+use lib::git::{AmendFastOptions, GitRunInfo, MaybeZeroOid, Repo, ResolvedReferenceInfo};
 
 /// Amends the existing HEAD commit.
 #[instrument]
@@ -26,20 +28,22 @@ pub fn amend(
     effects: &Effects,
     git_run_info: &GitRunInfo,
     move_options: &MoveOptions,
-) -> eyre::Result<isize> {
+) -> eyre::Result<ExitCode> {
     let now = SystemTime::now();
+    let timestamp = now.duration_since(SystemTime::UNIX_EPOCH)?.as_secs_f64();
     let repo = Repo::from_current_dir()?;
     let conn = repo.get_db_conn()?;
-    let mut event_log_db = EventLogDb::new(&conn)?;
+    let event_log_db = EventLogDb::new(&conn)?;
 
-    let head_oid = match repo.get_head_info()?.oid {
+    let head_info = repo.get_head_info()?;
+    let head_oid = match head_info.oid {
         Some(oid) => oid,
         None => {
             writeln!(
                 effects.get_output_stream(),
                 "No commit is currently checked out. Check out a commit to amend and then try again.",
             )?;
-            return Ok(1);
+            return Ok(ExitCode(1));
         }
     };
     let head_commit = repo.find_commit_or_fail(head_oid)?;
@@ -50,48 +54,59 @@ pub fn amend(
             effects.get_output_stream(),
             "Cannot amend, because there are unresolved merge conflicts. Resolve the merge conflicts and try again."
         )?;
-        return Ok(1);
+        return Ok(ExitCode(1));
     }
 
     let event_tx_id = event_log_db.make_transaction_id(now, "amend")?;
-    let staged_index_paths = repo.get_staged_paths()?;
-    let (opts, dirty_working_tree) = if !staged_index_paths.is_empty() {
-        let dirty_working_tree = repo.has_changed_files(effects, git_run_info)?;
-        let opts = AmendFastOptions::FromIndex {
-            paths: staged_index_paths.into_iter().collect_vec(),
-        };
-        (opts, dirty_working_tree)
+    let (snapshot, status) =
+        repo.get_status(effects, git_run_info, &index, &head_info, Some(event_tx_id))?;
+    {
+        let ResolvedReferenceInfo {
+            oid,
+            reference_name,
+        } = head_info;
+        event_log_db.add_events(vec![Event::WorkingCopySnapshot {
+            timestamp,
+            event_tx_id,
+            head_oid: MaybeZeroOid::from(oid),
+            commit_oid: snapshot.base_commit.get_oid(),
+            ref_name: reference_name.map(|name| name.into_owned()),
+        }])?;
+    }
+
+    // Note that there may be paths which are in both of these entries in the
+    // case that the given path has both staged and unstaged changes.
+    let staged_entries = status
+        .clone()
+        .into_iter()
+        .filter(|entry| entry.index_status.is_changed())
+        .collect_vec();
+    let unstaged_entries = status
+        .into_iter()
+        .filter(|entry| entry.working_copy_status.is_changed())
+        .collect_vec();
+
+    let opts = if !staged_entries.is_empty() {
+        AmendFastOptions::FromIndex {
+            paths: staged_entries
+                .into_iter()
+                .flat_map(|entry| entry.paths())
+                .collect(),
+        }
     } else {
-        let status = repo.get_status(git_run_info, Some(event_tx_id))?;
-        let entries_to_amend = status
-            .into_iter()
-            .filter(|entry| match entry.working_copy_status {
-                FileStatus::Added
-                | FileStatus::Copied
-                | FileStatus::Deleted
-                | FileStatus::Modified
-                | FileStatus::Renamed => true,
-                FileStatus::Ignored
-                | FileStatus::Unmerged
-                | FileStatus::Unmodified
-                | FileStatus::Untracked => false,
-            })
-            .collect_vec();
-        let opts = AmendFastOptions::FromWorkingCopy {
-            status_entries: entries_to_amend,
-        };
-        (opts, true)
+        AmendFastOptions::FromWorkingCopy {
+            status_entries: unstaged_entries.clone(),
+        }
     };
     if opts.is_empty() {
         writeln!(
             effects.get_output_stream(),
             "There are no uncommitted or staged changes. Nothing to amend."
         )?;
-        return Ok(0);
+        return Ok(ExitCode(0));
     }
 
     let amended_tree = repo.amend_fast(&head_commit, &opts)?;
-    let timestamp = now.duration_since(SystemTime::UNIX_EPOCH)?.as_secs_f64();
 
     let (author, committer) = (head_commit.get_author(), head_commit.get_committer());
     let (author, committer) = if get_restack_preserve_timestamps(&repo)? {
@@ -122,16 +137,20 @@ pub fn amend(
 
     if let AmendFastOptions::FromWorkingCopy { .. } = opts {
         // TODO(#201): Figure out a way to perform "fast amend" on the working copy without needing a reset.
-        git_run_info.run(effects, Some(event_tx_id), &["reset"])?;
+        let exit_code = git_run_info.run(effects, Some(event_tx_id), &["reset"])?;
+        if !exit_code.is_success() {
+            return Ok(exit_code);
+        }
     }
 
     let restack_exit_code = restack::restack(
         effects,
         git_run_info,
-        vec![head_oid.to_string()],
+        vec![Revset(head_oid.to_string())],
         move_options,
+        MergeConflictRemediation::Restack,
     )?;
-    if restack_exit_code != 0 {
+    if !restack_exit_code.is_success() {
         return Ok(restack_exit_code);
     }
 
@@ -144,7 +163,7 @@ pub fn amend(
             };
             let mut message = format!("Amended with {}.", staged_changes);
             // TODO: Include the number of uncommitted changes.
-            if dirty_working_tree {
+            if !unstaged_entries.is_empty() {
                 message += " (Some uncommitted changes were not amended.)";
             }
             writeln!(effects.get_output_stream(), "{}", message)?;
@@ -162,5 +181,5 @@ pub fn amend(
             )?;
         }
     }
-    Ok(0)
+    Ok(ExitCode(0))
 }
