@@ -50,7 +50,7 @@ pub fn hook_post_checkout(
 
     let repo = Repo::from_current_dir()?;
     let conn = repo.get_db_conn()?;
-    let mut event_log_db = EventLogDb::new(&conn)?;
+    let event_log_db = EventLogDb::new(&conn)?;
     let event_tx_id = event_log_db.make_transaction_id(now, "hook-post-checkout")?;
     event_log_db.add_events(vec![Event::RefUpdateEvent {
         timestamp: timestamp.as_secs_f64(),
@@ -71,7 +71,7 @@ fn hook_post_commit_common(effects: &Effects, hook_name: &str) -> eyre::Result<(
     let glyphs = Glyphs::detect();
     let repo = Repo::from_current_dir()?;
     let conn = repo.get_db_conn()?;
-    let mut event_log_db = EventLogDb::new(&conn)?;
+    let event_log_db = EventLogDb::new(&conn)?;
 
     let commit_oid = match repo.get_head_info()?.oid {
         Some(commit_oid) => commit_oid,
@@ -91,15 +91,16 @@ fn hook_post_commit_common(effects: &Effects, hook_name: &str) -> eyre::Result<(
     mark_commit_reachable(&repo, commit_oid)
         .wrap_err("Marking commit as reachable for GC purposes")?;
 
-    let timestamp = commit.get_time().seconds();
+    let timestamp = commit.get_time().to_system_time()?;
 
     // Potentially lossy conversion. The semantics are to round to the nearest
     // possible float:
     // https://doc.rust-lang.org/reference/expressions/operator-expr.html#semantics.
     // We don't rely on the timestamp's correctness for anything, so this is
     // okay.
-    #[allow(clippy::as_conversions)]
-    let timestamp = timestamp as f64;
+    let timestamp = timestamp
+        .duration_since(SystemTime::UNIX_EPOCH)?
+        .as_secs_f64();
 
     let event_tx_id = event_log_db.make_transaction_id(now, hook_name)?;
     event_log_db.add_events(vec![Event::CommitEvent {
@@ -407,7 +408,7 @@ pub fn hook_reference_transaction(effects: &Effects, transaction_state: &str) ->
 
     let repo = Repo::from_current_dir()?;
     let conn = repo.get_db_conn()?;
-    let mut event_log_db = EventLogDb::new(&conn)?;
+    let event_log_db = EventLogDb::new(&conn)?;
     let event_tx_id = event_log_db.make_transaction_id(now, "reference-transaction")?;
 
     let packed_references = read_packed_refs_file(&repo)?;
