@@ -9,13 +9,16 @@ use std::time::SystemTime;
 use cursive::theme::BaseColor;
 use cursive::utils::markup::StyledString;
 use eden_dag::DagAlgorithm;
+use lib::core::check_out::{check_out_commit, CheckOutCommitOptions};
+use lib::core::repo_ext::RepoExt;
+use lib::util::ExitCode;
 use tracing::{instrument, warn};
 
 use crate::commands::smartlog::make_smartlog_graph;
 use crate::opts::{CheckoutOptions, TraverseCommitsOptions};
 use crate::tui::prompt_select_commit;
 use lib::core::config::get_next_interactive;
-use lib::core::dag::{sort_commit_set, CommitSet, Dag};
+use lib::core::dag::{sorted_commit_set, CommitSet, Dag};
 use lib::core::effects::Effects;
 use lib::core::eventlog::{EventLogDb, EventReplayer};
 use lib::core::formatting::{printable_styled_string, Pluralize};
@@ -23,7 +26,7 @@ use lib::core::node_descriptors::{
     BranchesDescriptor, CommitMessageDescriptor, CommitOidDescriptor,
     DifferentialRevisionDescriptor, NodeDescriptor, Redactor, RelativeTimeDescriptor,
 };
-use lib::git::{check_out_commit, CheckOutCommitOptions, GitRunInfo, NonZeroOid, Repo};
+use lib::git::{GitRunInfo, NonZeroOid, Repo};
 
 /// The command being invoked, indicating which direction to traverse commits.
 #[derive(Clone, Copy, Debug)]
@@ -134,7 +137,7 @@ fn advance(
                     } => descendant_branches()?,
                 };
 
-                sort_commit_set(repo, dag, &children)?
+                sorted_commit_set(repo, dag, &children)?
             }
 
             Command::Prev => {
@@ -182,7 +185,7 @@ fn advance(
                     } => ancestor_branches()?,
                 };
 
-                sort_commit_set(repo, dag, &parents)?
+                sorted_commit_set(repo, dag, &parents)?
             }
         };
 
@@ -302,7 +305,7 @@ pub fn traverse_commits(
     git_run_info: &GitRunInfo,
     command: Command,
     options: &TraverseCommitsOptions,
-) -> eyre::Result<isize> {
+) -> eyre::Result<ExitCode> {
     let TraverseCommitsOptions {
         num_commits,
         all_the_way,
@@ -342,11 +345,19 @@ pub fn traverse_commits(
         }
     };
 
+    let now = SystemTime::now();
     let repo = Repo::from_current_dir()?;
     let head_info = repo.get_head_info()?;
     let references_snapshot = repo.get_references_snapshot()?;
     let conn = repo.get_db_conn()?;
     let event_log_db = EventLogDb::new(&conn)?;
+    let event_tx_id = event_log_db.make_transaction_id(
+        now,
+        match command {
+            Command::Next => "next",
+            Command::Prev => "prev",
+        },
+    )?;
     let event_replayer = EventReplayer::from_event_log_db(effects, &repo, &event_log_db)?;
     let event_cursor = event_replayer.make_default_cursor();
     let dag = Dag::open_and_sync(
@@ -386,7 +397,7 @@ pub fn traverse_commits(
         towards,
     )?;
     let current_oid = match current_oid {
-        None => return Ok(1),
+        None => return Ok(ExitCode(1)),
         Some(current_oid) => current_oid,
     };
 
@@ -426,22 +437,24 @@ pub fn traverse_commits(
     };
 
     let additional_args = {
-        let mut args = Vec::new();
+        let mut args: Vec<OsString> = Vec::new();
         if merge {
-            args.push("--merge");
+            args.push("--merge".into());
         }
         if force {
-            args.push("--force")
+            args.push("--force".into())
         }
         args
     };
     check_out_commit(
         effects,
         git_run_info,
-        None,
+        &repo,
+        &event_log_db,
+        event_tx_id,
         Some(&current_oid),
         &CheckOutCommitOptions {
-            additional_args: additional_args.as_slice(),
+            additional_args,
             ..Default::default()
         },
     )
@@ -516,7 +529,7 @@ pub fn checkout(
     effects: &Effects,
     git_run_info: &GitRunInfo,
     checkout_options: &CheckoutOptions,
-) -> eyre::Result<isize> {
+) -> eyre::Result<ExitCode> {
     let CheckoutOptions {
         interactive: _,
         branch_name,
@@ -525,11 +538,13 @@ pub fn checkout(
         target,
     } = checkout_options;
 
+    let now = SystemTime::now();
     let repo = Repo::from_current_dir()?;
     let head_info = repo.get_head_info()?;
     let references_snapshot = repo.get_references_snapshot()?;
     let conn = repo.get_db_conn()?;
     let event_log_db = EventLogDb::new(&conn)?;
+    let event_tx_id = event_log_db.make_transaction_id(now, "checkout")?;
     let event_replayer = EventReplayer::from_event_log_db(effects, &repo, &event_log_db)?;
     let event_cursor = event_replayer.make_default_cursor();
     let dag = Dag::open_and_sync(
@@ -546,8 +561,8 @@ pub fn checkout(
         &dag,
         &event_replayer,
         event_cursor,
+        &dag.observed_commits,
         true,
-        false,
     )?;
 
     let initial_query = get_initial_query(checkout_options);
@@ -572,22 +587,22 @@ pub fn checkout(
                 ],
             )? {
                 Some(oid) => Some(oid.to_string()),
-                None => return Ok(1),
+                None => return Ok(ExitCode(1)),
             }
         }
     };
 
     let additional_args = {
-        let mut args = Vec::new();
+        let mut args: Vec<OsString> = Vec::new();
         if let Some(branch_name) = branch_name {
-            args.push("-b");
-            args.push(branch_name);
+            args.push("-b".into());
+            args.push(branch_name.into());
         }
         if *force {
-            args.push("-f");
+            args.push("-f".into());
         }
         if *merge {
-            args.push("-m");
+            args.push("-m".into());
         }
         args
     };
@@ -595,10 +610,12 @@ pub fn checkout(
     let exit_code = check_out_commit(
         effects,
         git_run_info,
-        None,
+        &repo,
+        &event_log_db,
+        event_tx_id,
         target.as_ref(),
         &CheckOutCommitOptions {
-            additional_args: additional_args.as_slice(),
+            additional_args,
             render_smartlog: true,
         },
     )?;
