@@ -3,15 +3,22 @@
 //! The set of commits that are still being worked on is inferred from the event
 //! log; see the `eventlog` module.
 
+use std::cmp::Ordering;
 use std::fmt::Write;
+use std::mem::swap;
 use std::time::SystemTime;
 
+use console::style;
+use lib::core::config::{get_hint_enabled, print_hint_suppression_notice, Hint};
+use lib::core::repo_ext::RepoExt;
+use lib::core::rewrite::find_rewrite_target;
+use lib::util::ExitCode;
 use tracing::instrument;
 
-use lib::core::dag::Dag;
+use lib::core::dag::{CommitSet, Dag};
 use lib::core::effects::Effects;
 use lib::core::eventlog::{EventLogDb, EventReplayer};
-use lib::core::formatting::printable_styled_string;
+use lib::core::formatting::{printable_styled_string, Pluralize};
 use lib::core::node_descriptors::{
     BranchesDescriptor, CommitMessageDescriptor, CommitOidDescriptor,
     DifferentialRevisionDescriptor, ObsolescenceExplanationDescriptor, Redactor,
@@ -22,19 +29,21 @@ use lib::git::{GitRunInfo, Repo};
 pub use graph::{make_smartlog_graph, SmartlogGraph};
 pub use render::{render_graph, SmartlogOptions};
 
+use crate::revset::resolve_commits;
+
 mod graph {
     use std::collections::HashMap;
     use std::convert::TryFrom;
-    use std::ops::Deref;
 
     use eden_dag::DagAlgorithm;
+    use lib::core::gc::mark_commit_reachable;
     use tracing::instrument;
 
-    use lib::core::dag::{commit_set_to_vec, CommitSet, Dag};
+    use lib::core::dag::{commit_set_to_vec_unsorted, CommitSet, Dag};
     use lib::core::effects::{Effects, OperationType};
     use lib::core::eventlog::{EventCursor, EventReplayer};
     use lib::core::node_descriptors::NodeObject;
-    use lib::git::Commit;
+    use lib::git::{Commit, Time};
     use lib::git::{NonZeroOid, Repo};
 
     /// Node contained in the smartlog commit graph.
@@ -45,7 +54,7 @@ mod graph {
 
         /// The OID of the parent node in the smartlog commit graph.
         ///
-        /// This is different from inspecting `commit.parents()`, since the smartlog
+        /// This is different from inspecting `commit.parents()`,& since the smartlog
         /// will hide most nodes from the commit graph, including parent nodes.
         pub parent: Option<NonZeroOid>,
 
@@ -77,7 +86,7 @@ mod graph {
 
     /// Graph of commits that the user is working on.
     pub struct SmartlogGraph<'repo> {
-        nodes: HashMap<NonZeroOid, Node<'repo>>,
+        pub nodes: HashMap<NonZeroOid, Node<'repo>>,
     }
 
     impl<'repo> SmartlogGraph<'repo> {
@@ -104,14 +113,6 @@ mod graph {
         }
     }
 
-    impl<'repo> Deref for SmartlogGraph<'repo> {
-        type Target = HashMap<NonZeroOid, Node<'repo>>;
-
-        fn deref(&self) -> &Self::Target {
-            &self.nodes
-        }
-    }
-
     /// Find additional commits that should be displayed.
     ///
     /// For example, if you check out a commit that has intermediate parent commits
@@ -123,8 +124,6 @@ mod graph {
         effects: &Effects,
         repo: &'repo Repo,
         dag: &Dag,
-        event_replayer: &EventReplayer,
-        event_cursor: EventCursor,
         public_commits: &CommitSet,
         active_heads: &CommitSet,
     ) -> eyre::Result<SmartlogGraph<'repo>> {
@@ -176,7 +175,7 @@ mod graph {
             let mut links = Vec::new();
             for child_oid in non_main_node_oids {
                 let parent_vertexes = dag.query().parents(CommitSet::from(*child_oid))?;
-                let parent_oids = commit_set_to_vec(&parent_vertexes)?;
+                let parent_oids = commit_set_to_vec_unsorted(&parent_vertexes)?;
                 for parent_oid in parent_oids {
                     if graph.contains_key(&parent_oid) {
                         links.push((*child_oid, parent_oid))
@@ -197,7 +196,8 @@ mod graph {
     /// Sort children nodes of the commit graph in a standard order, for determinism
     /// in output.
     fn sort_children(graph: &mut SmartlogGraph) {
-        let commit_times: HashMap<NonZeroOid, Option<git2::Time>> = graph
+        let commit_times: HashMap<NonZeroOid, Option<Time>> = graph
+            .nodes
             .iter()
             .map(|(oid, node)| {
                 (
@@ -211,7 +211,7 @@ mod graph {
             .collect();
         for node in graph.nodes.values_mut() {
             node.children
-                .sort_by_key(|child_oid| (commit_times[child_oid], child_oid.to_string()));
+                .sort_by_key(|child_oid| (&commit_times[child_oid], child_oid.to_string()));
         }
     }
 
@@ -223,8 +223,8 @@ mod graph {
         dag: &Dag,
         event_replayer: &EventReplayer,
         event_cursor: EventCursor,
+        observed_commits: &CommitSet,
         remove_commits: bool,
-        only_branches: bool,
     ) -> eyre::Result<SmartlogGraph<'repo>> {
         let (effects, _progress) = effects.start_operation(OperationType::MakeGraph);
 
@@ -233,25 +233,18 @@ mod graph {
 
             let public_commits = dag.query_public_commits()?;
 
-            let observed_commits = if only_branches {
-                dag.branch_commits.clone()
-            } else if remove_commits {
-                dag.observed_commits.difference(&dag.obsolete_commits)
+            let observed_commits = if remove_commits {
+                observed_commits.difference(&dag.obsolete_commits)
             } else {
-                dag.observed_commits.clone()
+                observed_commits.clone()
             };
 
             let active_heads = dag.query_active_heads(&public_commits, &observed_commits)?;
+            for oid in commit_set_to_vec_unsorted(&active_heads)? {
+                mark_commit_reachable(repo, oid)?;
+            }
 
-            walk_from_active_heads(
-                &effects,
-                repo,
-                dag,
-                event_replayer,
-                event_cursor,
-                &public_commits,
-                &active_heads,
-            )?
+            walk_from_active_heads(&effects, repo, dag, &public_commits, &active_heads)?
         };
         sort_children(&mut graph);
         Ok(graph)
@@ -273,6 +266,8 @@ mod render {
     use lib::core::node_descriptors::{render_node_descriptors, NodeDescriptor};
     use lib::git::{NonZeroOid, Repo};
 
+    use crate::opts::Revset;
+
     use super::graph::SmartlogGraph;
 
     /// Split fully-independent subgraphs into multiple graphs.
@@ -289,6 +284,7 @@ mod render {
         graph: &SmartlogGraph,
     ) -> Vec<NonZeroOid> {
         let mut root_commit_oids: Vec<NonZeroOid> = graph
+            .nodes
             .iter()
             .filter(|(_oid, node)| node.parent.is_none())
             .map(|(oid, _node)| oid)
@@ -339,7 +335,7 @@ mod render {
         current_oid: NonZeroOid,
         last_child_line_char: Option<&str>,
     ) -> eyre::Result<Vec<StyledString>> {
-        let current_node = &graph[&current_oid];
+        let current_node = &graph.nodes[&current_oid];
         let is_head = Some(current_oid) == head_oid;
 
         let text = render_node_descriptors(glyphs, &current_node.object, commit_descriptors)?;
@@ -370,7 +366,7 @@ mod render {
         let children: Vec<_> = current_node
             .children
             .iter()
-            .filter(|child_oid| graph.contains_key(child_oid))
+            .filter(|child_oid| graph.nodes.contains_key(child_oid))
             .copied()
             .collect();
         for (child_idx, child_oid) in children.iter().enumerate() {
@@ -518,14 +514,30 @@ mod render {
     }
 
     /// Options for rendering the smartlog.
-    #[derive(Debug, Default)]
+    #[derive(Debug)]
     pub struct SmartlogOptions {
         /// Whether to also show commits in the smartlog which would normally not be
         /// visible.
         pub show_hidden_commits: bool,
 
-        /// Whether to only show commits on branches.
-        pub only_show_branches: bool,
+        /// The point in time at which to show the smartlog. If not provided,
+        /// renders the smartlog as of the current time. If negative, is treated
+        /// as an offset from the current event.
+        pub event_id: Option<isize>,
+
+        /// The commits to render. These commits and their ancestors up to the
+        /// main branch will be rendered.
+        pub revset: Revset,
+    }
+
+    impl Default for SmartlogOptions {
+        fn default() -> Self {
+            Self {
+                show_hidden_commits: Default::default(),
+                event_id: Default::default(),
+                revset: Revset("draft()".to_string()),
+            }
+        }
     }
 }
 
@@ -535,20 +547,34 @@ pub fn smartlog(
     effects: &Effects,
     git_run_info: &GitRunInfo,
     options: &SmartlogOptions,
-) -> eyre::Result<()> {
+) -> eyre::Result<ExitCode> {
     let SmartlogOptions {
         show_hidden_commits,
-        only_show_branches,
+        event_id,
+        revset,
     } = options;
 
     let repo = Repo::from_dir(&git_run_info.working_directory)?;
     let head_info = repo.get_head_info()?;
-    let references_snapshot = repo.get_references_snapshot()?;
     let conn = repo.get_db_conn()?;
     let event_log_db = EventLogDb::new(&conn)?;
     let event_replayer = EventReplayer::from_event_log_db(effects, &repo, &event_log_db)?;
-    let event_cursor = event_replayer.make_default_cursor();
-    let dag = Dag::open_and_sync(
+    let (references_snapshot, event_cursor) = {
+        let default_cursor = event_replayer.make_default_cursor();
+        match event_id {
+            None => (repo.get_references_snapshot()?, default_cursor),
+            Some(event_id) => {
+                let event_cursor = match event_id.cmp(&0) {
+                    Ordering::Less => event_replayer.advance_cursor(default_cursor, *event_id),
+                    Ordering::Equal | Ordering::Greater => event_replayer.make_cursor(*event_id),
+                };
+                let references_snapshot =
+                    event_replayer.get_references_snapshot(&repo, event_cursor)?;
+                (references_snapshot, event_cursor)
+            }
+        }
+    };
+    let mut dag = Dag::open_and_sync(
         effects,
         &repo,
         &event_replayer,
@@ -556,14 +582,44 @@ pub fn smartlog(
         &references_snapshot,
     )?;
 
+    let observed_commits = {
+        // For the purpose of resolving the revset expression, we may
+        // temporarily clear the DAG's obsolete commit set. However, when we
+        // render the smartlog later, we want to have the original obsolete
+        // commit set so that we can correctly identify which of the nodes are
+        // obsolete commits.
+        let mut old_obsolete_commits = CommitSet::empty();
+        if *show_hidden_commits {
+            swap(&mut dag.obsolete_commits, &mut old_obsolete_commits);
+        }
+        let observed_commits = match resolve_commits(effects, &repo, &mut dag, vec![revset.clone()])
+        {
+            Ok(result) => match result.as_slice() {
+                [commit_set] => commit_set.clone(),
+                other => panic!(
+                    "Expected exactly 1 result from resolve commits, got: {:?}",
+                    other
+                ),
+            },
+            Err(err) => {
+                err.describe(effects)?;
+                return Ok(ExitCode(1));
+            }
+        };
+        if *show_hidden_commits {
+            swap(&mut dag.obsolete_commits, &mut old_obsolete_commits);
+        }
+        observed_commits
+    };
+
     let graph = make_smartlog_graph(
         effects,
         &repo,
         &dag,
         &event_replayer,
         event_cursor,
+        &observed_commits,
         !show_hidden_commits,
-        *only_show_branches,
     )?;
 
     let lines = render_graph(
@@ -597,5 +653,39 @@ pub fn smartlog(
         )?;
     }
 
-    Ok(())
+    if !show_hidden_commits && get_hint_enabled(&repo, Hint::SmartlogFixAbandoned)? {
+        let num_abandoned_children = graph
+            .nodes
+            .iter()
+            .filter_map(|(oid, node)| {
+                if node.is_obsolete
+                    && find_rewrite_target(&event_replayer, event_cursor, *oid).is_some()
+                {
+                    Some(node.children.iter())
+                } else {
+                    None
+                }
+            })
+            .count();
+        if num_abandoned_children > 0 {
+            writeln!(
+                effects.get_output_stream(),
+                "{}: there {} in your commit graph",
+                style("hint").blue().bold(),
+                Pluralize {
+                    determiner: Some(("is", "are")),
+                    amount: num_abandoned_children,
+                    unit: ("abandoned commit", "abandoned commits"),
+                },
+            )?;
+            writeln!(
+                effects.get_output_stream(),
+                "{}: to fix this, run: git restack",
+                style("hint").blue().bold(),
+            )?;
+            print_hint_suppression_notice(effects, Hint::SmartlogFixAbandoned)?;
+        }
+    }
+
+    Ok(ExitCode(0))
 }
