@@ -1,8 +1,23 @@
 //! The command-line options for `git-branchless`.
 
 use clap::{ArgEnum, Args, Command as ClapCommand, IntoApp, Parser};
+use lib::git::NonZeroOid;
 use man::Arg;
 use std::path::{Path, PathBuf};
+use std::str::FromStr;
+
+/// A revset expression. Can be a commit hash, branch name, or one of the
+/// various revset functions.
+#[derive(Clone, Debug)]
+pub struct Revset(pub String);
+
+impl FromStr for Revset {
+    type Err = std::convert::Infallible;
+
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        Ok(Self(s.to_string()))
+    }
+}
 
 /// A command wrapped by `git-branchless wrap`. The arguments are forwarded to
 /// `git`.
@@ -16,35 +31,40 @@ pub enum WrappedCommand {
 /// Options for moving commits.
 #[derive(Args, Debug)]
 pub struct MoveOptions {
+    /// Force moving public commits, even though other people may have access to
+    /// those commits.
+    #[clap(action, short = 'f', long = "force-rewrite", visible_alias = "fr")]
+    pub force_rewrite_public_commits: bool,
+
     /// Only attempt to perform an in-memory rebase. If it fails, do not
     /// attempt an on-disk rebase.
-    #[clap(long = "in-memory", conflicts_with_all(&["force-on-disk", "merge"]))]
+    #[clap(action, long = "in-memory", conflicts_with_all(&["force-on-disk", "merge"]))]
     pub force_in_memory: bool,
 
     /// Skip attempting to use an in-memory rebase, and try an
     /// on-disk rebase directly.
-    #[clap(long = "on-disk")]
+    #[clap(action, long = "on-disk")]
     pub force_on_disk: bool,
 
     /// Don't attempt to deduplicate commits. Normally, a commit with the same
     /// contents as another commit which has already been applied to the target
     /// branch is skipped. If set, this flag skips that check.
-    #[clap(long = "no-deduplicate-commits", parse(from_flag = std::ops::Not::not))]
+    #[clap(action(clap::ArgAction::SetFalse), long = "no-deduplicate-commits")]
     pub detect_duplicate_commits_via_patch_id: bool,
 
     /// Attempt to resolve merge conflicts, if any. If a merge conflict
     /// occurs and this option is not set, the operation is aborted.
-    #[clap(name = "merge", short = 'm', long = "merge")]
+    #[clap(action, name = "merge", short = 'm', long = "merge")]
     pub resolve_merge_conflicts: bool,
 
     /// Debugging option. Print the constraints used to create the rebase
     /// plan before executing it.
-    #[clap(long = "debug-dump-rebase-constraints")]
+    #[clap(action, long = "debug-dump-rebase-constraints")]
     pub dump_rebase_constraints: bool,
 
     /// Debugging option. Print the rebase plan that will be executed before
     /// executing it.
-    #[clap(long = "debug-dump-rebase-plan")]
+    #[clap(action, long = "debug-dump-rebase-plan")]
     pub dump_rebase_plan: bool,
 }
 
@@ -54,27 +74,29 @@ pub struct TraverseCommitsOptions {
     /// The number of commits to traverse.
     ///
     /// If not provided, defaults to 1.
+    #[clap(value_parser)]
     pub num_commits: Option<usize>,
 
     /// Traverse as many commits as possible.
-    #[clap(short = 'a', long = "all")]
+    #[clap(action, short = 'a', long = "all")]
     pub all_the_way: bool,
 
     /// Move the specified number of branches rather than commits.
-    #[clap(short = 'b', long = "branch")]
+    #[clap(action, short = 'b', long = "branch")]
     pub move_by_branches: bool,
 
     /// When encountering multiple next commits, choose the oldest.
-    #[clap(short = 'o', long = "oldest")]
+    #[clap(action, short = 'o', long = "oldest")]
     pub oldest: bool,
 
     /// When encountering multiple next commits, choose the newest.
-    #[clap(short = 'n', long = "newest", conflicts_with("oldest"))]
+    #[clap(action, short = 'n', long = "newest", conflicts_with("oldest"))]
     pub newest: bool,
 
     /// When encountering multiple next commits, interactively prompt which to
     /// advance to.
     #[clap(
+        action,
         short = 'i',
         long = "interactive",
         conflicts_with("newest"),
@@ -84,12 +106,12 @@ pub struct TraverseCommitsOptions {
 
     /// If the local changes conflict with the destination commit, attempt to
     /// merge them.
-    #[clap(short = 'm', long = "merge")]
+    #[clap(action, short = 'm', long = "merge")]
     pub merge: bool,
 
     /// If the local changes conflict with the destination commit, discard them.
     /// (Use with caution!)
-    #[clap(short = 'f', long = "force", conflicts_with("merge"))]
+    #[clap(action, short = 'f', long = "force", conflicts_with("merge"))]
     pub force: bool,
 }
 
@@ -97,22 +119,22 @@ pub struct TraverseCommitsOptions {
 #[derive(Args, Debug)]
 pub struct CheckoutOptions {
     /// Interactively select a commit to check out.
-    #[clap(short = 'i', long = "interactive")]
+    #[clap(action, short = 'i', long = "interactive")]
     pub interactive: bool,
 
     /// When checking out the target commit, also create a branch with the
     /// provided name pointing to that commit.
-    #[clap(short = 'b', long = "branch")]
+    #[clap(value_parser, short = 'b', long = "branch")]
     pub branch_name: Option<String>,
 
     /// Forcibly switch commits, discarding any working copy changes if
     /// necessary.
-    #[clap(short = 'f', long = "force")]
+    #[clap(action, short = 'f', long = "force")]
     pub force: bool,
 
     /// If the current working copy changes do not apply cleanly to the
     /// target commit, start merge conflict resolution instead of aborting.
-    #[clap(short = 'm', long = "merge", conflicts_with("force"))]
+    #[clap(action, short = 'm', long = "merge", conflicts_with("force"))]
     pub merge: bool,
 
     /// The commit or branch to check out.
@@ -122,6 +144,7 @@ pub struct CheckoutOptions {
     ///
     /// If this is provided and the `--interactive` flag is passed, this
     /// text is used to pre-fill the interactive commit selector.
+    #[clap(value_parser)]
     pub target: Option<String>,
 }
 
@@ -152,78 +175,98 @@ pub enum Command {
     /// Hide the provided commits from the smartlog.
     Hide {
         /// Zero or more commits to hide.
-        ///
-        /// Can either be hashes, like `abc123`, or ref-specs, like `HEAD^`.
-        commits: Vec<String>,
+        #[clap(value_parser)]
+        revsets: Vec<Revset>,
+
+        /// Also delete any branches that are abandoned as a result of this hide.
+        #[clap(action, short = 'D', long = "delete-branches")]
+        delete_branches: bool,
 
         /// Also recursively hide all visible children commits of the provided
         /// commits.
-        #[clap(short = 'r', long = "recursive")]
+        #[clap(action, short = 'r', long = "recursive")]
         recursive: bool,
     },
 
     /// Internal use.
+    #[clap(hide = true)]
     HookDetectEmptyCommit {
         /// The OID of the commit currently being applied, to be checked for emptiness.
+        #[clap(value_parser)]
         old_commit_oid: String,
     },
 
     /// Internal use.
+    #[clap(hide = true)]
     HookPreAutoGc,
 
     /// Internal use.
+    #[clap(hide = true)]
     HookPostCheckout {
         /// The previous commit OID.
+        #[clap(value_parser)]
         previous_commit: String,
 
         /// The current commit OID.
+        #[clap(value_parser)]
         current_commit: String,
 
         /// Whether or not this was a branch checkout (versus a file checkout).
+        #[clap(value_parser)]
         is_branch_checkout: isize,
     },
 
     /// Internal use.
+    #[clap(hide = true)]
     HookPostCommit,
 
     /// Internal use.
+    #[clap(hide = true)]
     HookPostMerge {
         /// Whether or not this is a squash merge. See githooks(5).
+        #[clap(value_parser)]
         is_squash_merge: isize,
     },
 
     /// Internal use.
+    #[clap(hide = true)]
     HookPostRewrite {
         /// One of `amend` or `rebase`.
+        #[clap(value_parser)]
         rewrite_type: String,
     },
 
     /// Internal use.
+    #[clap(hide = true)]
     HookReferenceTransaction {
         /// One of `prepared`, `committed`, or `aborted`. See githooks(5).
+        #[clap(value_parser)]
         transaction_state: String,
     },
 
     /// Internal use.
+    #[clap(hide = true)]
     HookRegisterExtraPostRewriteHook,
 
     /// Internal use.
+    #[clap(hide = true)]
     HookSkipUpstreamAppliedCommit {
         /// The OID of the commit that was skipped.
+        #[clap(value_parser)]
         commit_oid: String,
     },
 
     /// Initialize the branchless workflow for this repository.
     Init {
         /// Uninstall the branchless workflow instead of initializing it.
-        #[clap(long = "uninstall")]
+        #[clap(action, long = "uninstall")]
         uninstall: bool,
 
         /// Use the provided name as the name of the main branch.
         ///
         /// If not set, it will be auto-detected. If it can't be auto-detected,
         /// then you will be prompted to enter a value for the main branch name.
-        #[clap(long = "main-branch", conflicts_with = "uninstall")]
+        #[clap(value_parser, long = "main-branch", conflicts_with = "uninstall")]
         main_branch_name: Option<String>,
     },
 
@@ -239,19 +282,40 @@ pub enum Command {
     Move {
         /// The source commit to move. This commit, and all of its descendants,
         /// will be moved.
-        #[clap(short = 's', long = "source")]
-        source: Option<String>,
+        #[clap(action(clap::ArgAction::Append), short = 's', long = "source")]
+        source: Vec<Revset>,
 
         /// A commit inside a subtree to move. The entire subtree, starting from
         /// the main branch, will be moved, not just the commits descending from
         /// this commit.
-        #[clap(short = 'b', long = "base", conflicts_with = "source")]
-        base: Option<String>,
+        #[clap(
+            action(clap::ArgAction::Append),
+            short = 'b',
+            long = "base",
+            conflicts_with = "source"
+        )]
+        base: Vec<Revset>,
+
+        /// A set of specific commits to move. These will be removed from their
+        /// current locations and any unmoved children will be moved to their
+        /// nearest unmoved ancestor.
+        #[clap(
+            action(clap::ArgAction::Append),
+            short = 'x',
+            long = "exact",
+            conflicts_with_all(&["source", "base"])
+        )]
+        exact: Vec<Revset>,
 
         /// The destination commit to move all source commits onto. If not
         /// provided, defaults to the current commit.
-        #[clap(short = 'd', long = "dest")]
-        dest: Option<String>,
+        #[clap(value_parser, short = 'd', long = "dest")]
+        dest: Option<Revset>,
+
+        /// Insert the subtree between the destination and it's children, if any.
+        /// Only supported if the moved subtree has a single head.
+        #[clap(action, short = 'I', long = "insert")]
+        insert: bool,
 
         /// Options for moving commits.
         #[clap(flatten)]
@@ -272,46 +336,105 @@ pub enum Command {
         traverse_commits_options: TraverseCommitsOptions,
     },
 
+    /// Query the commit graph using the "revset" DSL.
+    Query {
+        /// The query to execute.
+        #[clap(value_parser)]
+        revset: Revset,
+
+        /// Print the branches attached to the resulting commits, rather than the commits themselves.
+        #[clap(action, short = 'b', long = "branches")]
+        show_branches: bool,
+
+        /// Print the OID of each matching commit, one per line. This output is
+        /// stable for use in scripts.
+        #[clap(action, short = 'r', long = "raw", conflicts_with("show-branches"))]
+        raw: bool,
+    },
+
+    /// Restore internal invariants by reconciling the internal operation log
+    /// with the state of the Git repository.
+    Repair {
+        /// Apply changes.
+        #[clap(action(clap::ArgAction::SetFalse), long = "no-dry-run")]
+        dry_run: bool,
+    },
+
     /// Fix up commits abandoned by a previous rewrite operation.
     Restack {
         /// The IDs of the abandoned commits whose descendants should be
         /// restacked. If not provided, all abandoned commits are restacked.
-        commits: Vec<String>,
+        #[clap(value_parser)]
+        commits: Vec<Revset>,
 
         /// Options for moving commits.
         #[clap(flatten)]
         move_options: MoveOptions,
     },
 
+    /// Create a commit by interactively selecting which changes to include.
+    Record {
+        /// The commit message to use. If not provided, will be prompted to provide a commit message
+        /// interactively.
+        #[clap(value_parser, short = 'm', long = "message")]
+        message: Option<String>,
+
+        /// Select changes to include interactively, rather than using the
+        /// current staged/unstaged changes.
+        #[clap(action, short = 'i', long = "interactive")]
+        interactive: bool,
+    },
+
     /// Reword commits.
     Reword {
         /// Zero or more commits to reword. If not provided, defaults to "HEAD".
         ///
         /// Can either be hashes, like `abc123`, or ref-specs, like `HEAD^`.
-        commits: Vec<String>,
+        #[clap(value_parser)]
+        revsets: Vec<Revset>,
+
+        /// Force rewording public commits, even though other people may have access to
+        /// those commits.
+        #[clap(action, short = 'f', long = "force-rewrite", visible_alias = "fr")]
+        force_rewrite_public_commits: bool,
 
         /// Message to apply to commits. Multiple messages will be combined as separate paragraphs,
         /// similar to `git commit`.
-        #[clap(short = 'm', long = "message")]
+        #[clap(value_parser, short = 'm', long = "message")]
         messages: Vec<String>,
 
         /// Throw away the original commit messages.
         ///
         /// If `commit.template` is set, then the editor is pre-populated with
         /// that; otherwise, the editor starts empty.
-        #[clap(short = 'd', long = "discard", conflicts_with("messages"))]
+        #[clap(action, short = 'd', long = "discard", conflicts_with("messages"))]
         discard: bool,
     },
 
     /// Display a nice graph of the commits you've recently worked on.
     Smartlog {
         /// Also show commits which have been hidden.
-        #[clap(long = "hidden")]
+        #[clap(action, long = "hidden")]
         show_hidden_commits: bool,
 
-        /// Only show commits that exist on a branch.
-        #[clap(long = "only-branches", conflicts_with = "show-hidden-commits")]
-        only_show_branches: bool,
+        /// The point in time at which to show the smartlog. If not provided,
+        /// renders the smartlog as of the current time. If negative, is treated
+        /// as an offset from the current event.
+        #[clap(value_parser, long = "event-id")]
+        event_id: Option<isize>,
+
+        /// The commits to render. These commits and their ancestors up to the
+        /// main branch will be rendered.
+        #[clap(value_parser, default_value = "draft()")]
+        revset: Revset,
+    },
+
+    #[clap(hide = true)]
+    /// Manage working copy snapshots.
+    Snapshot {
+        /// The subcommand to run.
+        #[clap(subcommand)]
+        subcommand: SnapshotSubcommand,
     },
 
     /// Move any local commit stacks on top of the main branch.
@@ -319,6 +442,7 @@ pub enum Command {
         /// Run `git fetch` to update remote references before carrying out the
         /// sync.
         #[clap(
+            action,
             short = 'p',
             long = "pull",
             visible_short_alias = 'u',
@@ -326,46 +450,43 @@ pub enum Command {
         )]
         update_refs: bool,
 
-        /// Force rebasing commits even if they're already based on top of their
-        /// destination.
-        #[clap(short = 'f', long = "force")]
-        force: bool,
-
         /// Options for moving commits.
         #[clap(flatten)]
         move_options: MoveOptions,
 
         /// The commits whose stacks will be moved on top of the main branch. If
         /// no commits are provided, all draft commits will be synced.
-        ///
-        /// Can either be hashes, like `abc123`, or ref-specs, like `HEAD^`.
-        commits: Vec<String>,
+        #[clap(value_parser)]
+        revsets: Vec<Revset>,
     },
 
     /// Browse or return to a previous state of the repository.
     Undo {
         /// Interactively browse through previous states of the repository
         /// before selecting one to return to.
-        #[clap(short = 'i', long = "interactive")]
+        #[clap(action, short = 'i', long = "interactive")]
         interactive: bool,
+
+        /// Skip confirmation and apply changes immediately.
+        #[clap(action, short = 'y', long = "yes")]
+        yes: bool,
     },
 
     /// Unhide previously-hidden commits from the smartlog.
     Unhide {
         /// Zero or more commits to unhide.
-        ///
-        /// Can either be hashes, like `abc123`, or ref-specs, like `HEAD^`.
-        commits: Vec<String>,
+        #[clap(value_parser)]
+        revsets: Vec<Revset>,
 
         /// Also recursively unhide all children commits of the provided commits.
-        #[clap(short = 'r', long = "recursive")]
+        #[clap(action, short = 'r', long = "recursive")]
         recursive: bool,
     },
 
     /// Wrap a Git command inside a branchless transaction.
     Wrap {
         /// The `git` executable to invoke.
-        #[clap(long = "git-executable")]
+        #[clap(value_parser, long = "git-executable")]
         git_executable: Option<PathBuf>,
 
         /// The arguments to pass to `git`.
@@ -394,11 +515,11 @@ pub enum ColorSetting {
 pub struct Opts {
     /// Change to the given directory before executing the rest of the program.
     /// (The option is called `-C` for symmetry with Git.)
-    #[clap(short = 'C')]
+    #[clap(value_parser, short = 'C')]
     pub working_directory: Option<PathBuf>,
 
     /// Flag to force enable or disable terminal colors.
-    #[clap(long = "color", arg_enum)]
+    #[clap(value_parser, long = "color", arg_enum)]
     pub color: Option<ColorSetting>,
 
     /// The `git-branchless` subcommand to run.
@@ -406,6 +527,23 @@ pub struct Opts {
     pub command: Command,
 }
 
+/// `snapshot` subcommands.
+#[derive(Parser)]
+pub enum SnapshotSubcommand {
+    /// Create a new snapshot containing the working copy contents, and then
+    /// reset the working copy to the current `HEAD` commit.
+    ///
+    /// On success, prints the snapshot commit hash to stdout.
+    Create,
+
+    /// Restore the working copy contents from the provided snapshot.
+    Restore {
+        /// The commit hash for the snapshot.
+        #[clap(value_parser)]
+        snapshot_oid: NonZeroOid,
+    },
+}
+
 /// Generate and write man-pages into the specified directory.
 ///
 /// The generated files are named things like `man1/git-branchless-smartlog.1`,
